import json
from pathlib import Path
from typing import Any, Dict

from fastapi import HTTPException, status

from ucsschool.apis.utils import get_logger
from ucsschool.kelvin.client import (
    KelvinClientError,
    KelvinObject,
    KelvinResource,
    NoObject,
    Session as _Session,
)

kelvin_session_kwargs: Dict[str, Any] = {}


class Session(_Session):
    async def request(self, *args, **kwargs) -> Any:
        # Need to overwrite because the timeout can't be set in the session (Bug #54462)
        kwargs.update({"timeout": 4 * 60})  # The OpenAPI client uses 5 * 60
        return await super().request(*args, **kwargs)


_kelvin_session: Session = None


def load_kelvin_session_kwargs(path: Path) -> Dict[str, str]:
    try:
        with path.open() as settings_file:
            json_settings = json.load(settings_file)
        # only return required settings
        return {
            "username": json_settings["username"],
            "password": json_settings["password"],
            "host": json_settings["host"],
            "verify": json_settings["verify_ssl"],
        }
    except (EnvironmentError, KeyError, ValueError) as exc:
        raise ValueError(f"Missing, incomplete or malformed settings file '{path!s}': {exc!s}") from exc


async def kelvin_session() -> Session:
    # this function must be async, or Session.__init__() will fail because there is not yet an event loop
    global _kelvin_session
    if not _kelvin_session:
        _kelvin_session = Session(**kelvin_session_kwargs)
        _kelvin_session.open()
    return _kelvin_session


async def get_kelvin_obj(resource: KelvinResource, name: str, school: str = None) -> KelvinObject:
    """
    Retrieve an object via the Kelvin API.
    This is much faster than searching, if `school` and `name` are known (for `SchoolResource` and
    `UserResource` the `school` argument can be omitted).

    :param resource: The KelvinResource to use for retrieving the object
    :param str school: the OU of the object
    :param str name: the name of the object
    :return: The found KelvinObject
    :raises HTTPException: If the request failed or no object was found
    """
    logger = get_logger()
    get_kwargs = {"name": name}
    if school is not None:
        get_kwargs["school"] = school
    try:
        return await resource.get(**get_kwargs)
    except NoObject:
        logger.debug("%s not found with %r.", resource.Meta.kelvin_object.__name__, get_kwargs)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"{resource.Meta.kelvin_object.__name__} not found.",
        )
    except KelvinClientError as exc:
        logger.error(
            "Received non-404 KelvinClientError when retrieving %s with %r: %s",
            resource.Meta.kelvin_object.__name__,
            get_kwargs,
            exc,
        )
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Internal error retrieving {resource.Meta.kelvin_object.__name__}.",
        )


async def search_kelvin_obj(resource: KelvinResource, **search_args) -> KelvinObject:
    """
    Searches for an object via the Kelvin API and returns it.

    :param resource: The KelvinResource to use for searching for the object
    :param search_args: search arguments passed to the resources search method
    :return: The found KelvinObject
    :raises HTTPException: If the request failed, if no object or if more than one object was found
    """
    logger = get_logger()
    try:
        kelvin_objects = [kelvin_object async for kelvin_object in resource.search(**search_args)]
    except KelvinClientError as exc:
        raise HTTPException(status_code=exc.status, detail=exc.reason) from exc
    if len(kelvin_objects) == 0:
        logger.debug(
            "%s not found with search params: %r.", resource.Meta.kelvin_object.__name__, search_args
        )
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"{resource.Meta.kelvin_object.__name__} not found.",
        )
    elif len(kelvin_objects) > 1:
        logger.error(
            "Multiple %s objects found with search params: %r",
            resource.Meta.kelvin_object.__name__,
            search_args,
        )
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"More than one {resource.Meta.kelvin_object.__name__} was found.",
        )
    return kelvin_objects[0]
