import logging
from typing import Dict, Set

from fastapi import Depends, HTTPException, status
from ldap3.utils.conv import escape_filter_chars

from id_broker_common.kelvin import get_kelvin_obj, kelvin_session
from id_broker_common.udm_rest import (
    IdBrokerPseudonymisationError,
    get_service_provider_attribute,
    ldap_credentials,
    ldap_settings,
)
from id_broker_common.utils import remove_prefix
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.utils import LDAPAccess, auth_manager, get_logger
from ucsschool.kelvin.client import SchoolClass, SchoolClassResource, Session, User, UserResource

from .models import ClassMembers, NonEmptyStr, NoStarStr, Student, Teacher

# Extracted to variable to allow dep overrides in FastAPI, which cannot resolve anonymous dependencies
# TODO: import PLUGIN_NAME without circular import error:
token_dep = auth_manager("self_disclosure", OAuth2Token)


def service_provider_from_token(token: OAuth2Token = Depends(token_dep)) -> str:
    return token.client_id


def connecting_user_pseudonym_from_token(token: OAuth2Token = Depends(token_dep)) -> str:
    return token.sub


def roles_in_school(user: User, school_name: str) -> Set[str]:
    """Helper to get only the roles that are defined for a user within a specific school"""
    roles = set()
    for role_string in user.ucsschool_roles:
        role, context_type, context_name = role_string.split(":")
        if context_type == "school" and school_name == context_name:
            roles.add(role)
    return roles


async def extract_members(
    schoolclass: SchoolClass,
    service_provider: str,
    session: Session,
    list_students=True,
    list_teachers=True,
) -> ClassMembers:
    """
    Helper to get users of a school_class split into teachers and students

    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
    loaded from / found in `settings/data` UDM object
    """
    teachers = []
    students = []
    user_name: str
    sp_attribute = await get_service_provider_attribute(service_provider)
    for user_name in schoolclass.users:
        user: User = await get_kelvin_obj(UserResource(session=session), name=user_name)
        school_authority_prefix = f"{user.source_uid}-"
        if sp_attribute not in user.udm_properties:
            logger = get_logger()
            logger.warning(
                f"Ignoring user {user_name!r} without pseudonym in {sp_attribute!r} (for service "
                f"provider {service_provider!r})."
            )
            continue
        roles = roles_in_school(user, schoolclass.school)
        if list_students and "student" in roles:
            student = Student(
                user_id=user.udm_properties[sp_attribute],
                username=remove_prefix(user.name, school_authority_prefix),
                firstname=user.firstname,
                lastname=user.lastname,
            )
            students.append(student)
        if list_teachers and "teacher" in roles:
            teacher = Teacher(
                user_id=user.udm_properties[sp_attribute],
                username=remove_prefix(user.name, school_authority_prefix),
            )
            teachers.append(teacher)
    return ClassMembers(students=students, teachers=teachers)


async def get_user(pseudonym: str, service_provider: str, session: Session) -> User:
    """
    Retrieve user.

    :param str pseudonym: a users pseudonym
    :param str service_provider: the service provider for which to search user with the given pseudonym
    :param Session session: Kelvin client session instance
    :return: Kelvin User object
    :raises HTTPException: 404 if not found
    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    sp_attribute = await get_service_provider_attribute(service_provider)
    search_params = {sp_attribute: pseudonym}
    async for user in UserResource(session=session).search(**search_params):
        return user
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No user was found with pseudonym {pseudonym!r} for service provider "
            f"{service_provider!r}.",
        )


async def get_connected_user(
    service_provider: str = Depends(service_provider_from_token),
    pseudonym_of_connecting_user: str = Depends(connecting_user_pseudonym_from_token),
    session: Session = Depends(kelvin_session),
    logger: logging.Logger = Depends(get_logger),
) -> User:
    logger.debug(
        "service_provider=%r pseudonym_of_connecting_user=%r",
        service_provider,
        pseudonym_of_connecting_user,
    )
    try:
        return await get_user(pseudonym_of_connecting_user, service_provider, session)
    except (HTTPException, IdBrokerPseudonymisationError):
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


async def get_school_class(pseudonym: str, service_provider: str, session: Session) -> SchoolClass:
    """
    Retrieve school class.

    :param str pseudonym: a school class' pseudonym
    :param str service_provider: the service provider for which to search group with the given pseudonym
    :param Session session: Kelvin client session instance
    :return: Kelvin SchoolClass object
    :raises HTTPException: 404 if not found
    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    sp_attribute = await get_service_provider_attribute(service_provider)
    attrs = await _get_obj_attr_by_pseudonym(pseudonym, sp_attribute, "ucsschoolGroup", "cn")
    school = attrs["school"]
    name = remove_prefix(attrs["name"], f"{school}-")
    return await get_kelvin_obj(
        SchoolClassResource(session=session),
        name=name,
        school=school,
    )


async def _get_obj_attr_by_pseudonym(
    pseudonym: NoStarStr,
    sp_attribute: str,
    object_class: NonEmptyStr,
    name_attr: NonEmptyStr,
) -> Dict[str, str]:
    """Get 'name' and 'school' from `name_attr` and `ucsschoolRole` of LDAP object."""
    ldap = LDAPAccess(ldap_settings(), ldap_credentials())
    filter_s = f"(&(objectClass={object_class})" f"({sp_attribute}={escape_filter_chars(pseudonym)}))"
    results = await ldap.search(filter_s, attributes=[name_attr, "ucsschoolRole"])
    if len(results) == 1:
        result = results[0]
        name = result[name_attr].value
        if isinstance(result.ucsschoolRole.value, str):
            school_value = result.ucsschoolRole.value
        else:
            school_value = [s for s in result.ucsschoolRole.value if ":school:" in s][0]
        school = school_value.split(":")[-1]
    elif len(results) > 1:
        logger = get_logger()
        logger.error("More than 1 result when searching LDAP with filter %r: %r.", filter_s, results)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Error retrieving object with pseudonym {pseudonym!r}.",
        )
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Object with pseudonym {pseudonym!r} not found.",
        )
    return {"name": name, "school": school}
