import datetime
from unittest.mock import AsyncMock

import pytest
from fastapi.testclient import TestClient

from id_broker_common.kelvin import kelvin_session
from self_disclosure_plugin import dependencies
from self_disclosure_plugin.routes.v1 import users as users_module
from self_disclosure_plugin.routes.v1.users import router as users_router
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.opa import opa_instance
from ucsschool.kelvin.client import SchoolClass

from ..utils import DictObject, get_preconfigured_session_mock


@pytest.fixture
def client(app, dependency_overrides, fake_opa, async_dict):
    def _func(router, sub: str, client_id: str):
        app.include_router(router)
        overrides = {
            opa_instance: fake_opa,
            dependencies.token_dep: lambda: OAuth2Token(
                iss="",
                sub=sub,
                aud=[],
                exp=datetime.datetime.now() + datetime.timedelta(hours=3),
                client_id=client_id,
                iat=datetime.datetime.now() - datetime.timedelta(seconds=1),
                jti="",
            ),
            kelvin_session: get_preconfigured_session_mock(),
            dependencies.connecting_user_pseudonym_from_token: lambda: sub,
            dependencies.service_provider_from_token: lambda: client_id,
        }
        dependency_overrides(overrides)
        return TestClient(app)

    return _func


traeger1 = "EinTraeger"
service_provider_attribute = "idBrokerPseudonym0001"

student1 = DictObject(
    __name__=f"{traeger1}-student1",  # needed for parametrize
    udm_properties={"idBrokerPseudonym0001": "stu1@pseudo"},
    name=f"{traeger1}-student1",
    roles=["student"],
    firstname="stu1",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class1"]},
    school=f"{traeger1}-demoschool",
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
student2 = DictObject(
    __name__=f"{traeger1}-student2",  # needed for parametrize
    udm_properties={"idBrokerPseudonym0001": "stu2@pseudo"},
    name=f"{traeger1}-student2",
    roles=["student"],
    firstname="stu2",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class1"]},
    school=f"{traeger1}-demoschool",
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
student3 = DictObject(
    __name__=f"{traeger1}-student3",  # needed for parametrize
    udm_properties={"idBrokerPseudonym0001": "stu3@pseudo"},
    name=f"{traeger1}-student3",
    roles=["student"],
    firstname="stu3",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class2"]},
    school=f"{traeger1}-demoschool",
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
student4 = DictObject(
    __name__=f"{traeger1}-student4",  # needed for parametrize
    udm_properties={},  # no pseudonym!
    name=f"{traeger1}-student4",
    roles=["student"],
    firstname="stu4",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class2"]},
    school=f"{traeger1}-demoschool",
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
teacher1 = DictObject(
    __name__=f"{traeger1}-teacher1",  # needed for parametrize
    udm_properties={"idBrokerPseudonym0001": "tea1@Pseudonym"},
    name=f"{traeger1}-teacher1",
    roles=["teacher"],
    firstname="tea1",
    lastname="cher",
    school_classes={f"{traeger1}-demoschool": ["class1"]},
    school=f"{traeger1}-demoschool",
    source_uid=traeger1,
    ucsschool_roles=[f"teacher:school:{traeger1}-demoschool"],
)
class1 = DictObject(
    name="class1",
    users=[
        student1.name,
        student2.name,
        teacher1.name,
    ],
    school=f"{traeger1}-demoschool",
    udm_properties={"idBrokerPseudonym0001": "class1@pseudo", "ucsschoolRecordUID": "class1@record"},
)
class2 = DictObject(
    name="class2",
    users=[student3.name, student4.name],
    school=f"{traeger1}-demoschool",
    udm_properties={"idBrokerPseudonym0001": "class2@pseudo", "ucsschoolRecordUID": "class2@record"},
)


async def get_kelvin_obj_mock(resource, name, school=None):
    return {
        student1.name: student1,
        student2.name: student2,
        student3.name: student3,
        student4.name: student4,
        teacher1.name: teacher1,
        class1.name: class1,
        class2.name: class2,
    }[name]


def test_users_metadata(client, monkeypatch):
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=student1))
    test_client = client(
        router=users_router,
        sub=student1.udm_properties["idBrokerPseudonym0001"],
        client_id="ServiceProvider",
    )
    response = test_client.get(f"/{student1.udm_properties['idBrokerPseudonym0001']}/metadata")
    assert response.status_code == 200, response.__dict__
    prefix = f"{student1.source_uid}-"
    expected_user = {
        "user_id": student1.udm_properties["idBrokerPseudonym0001"],
        "username": student1.name.removeprefix(prefix),
        "firstname": student1.firstname,
        "lastname": student1.lastname,
        "type": student1.roles[0],
        "school_id": student1.school.removeprefix(prefix),
    }
    assert response.json() == expected_user


def test_user_groups(client, monkeypatch):
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=student1))
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(dependencies, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(users_module, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(
        users_module,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    test_client = client(
        router=users_router,
        sub=teacher1.udm_properties["idBrokerPseudonym0001"],
        client_id="ServiceProvider",
    )
    response = test_client.get(f"/{teacher1.udm_properties['idBrokerPseudonym0001']}/groups")
    assert response.status_code == 200
    expected_groups = {
        "groups": [
            {
                "group_id": class1.udm_properties["idBrokerPseudonym0001"],
                "name": class1.name,
                "student_count": 2,
            }
        ]
    }
    assert response.json() == expected_groups


@pytest.mark.skip("Test does not work and need to be fixed when proper pseudonyms are implemented")
def test_group_users(client, monkeypatch):
    users = [
        teacher1,
        student1,
        student2,
    ]
    school_classes = class1
    monkeypatch.setattr(dependencies, "schoolclass_getter", AsyncMock(return_value=school_classes))
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(side_effect=users))
    response = client.get(f"/groups/{class1.school}-{class1.name}/users")
    assert response.status_code == 200
    expected_group = dict(
        students=[
            dict(
                user_id=student1.udm_properties["brokerID"],
                username=student1.name,
                firstname=student1.firstname,
                lastname=student1.lastname,
            ),
            dict(
                user_id=student2.udm_properties["brokerID"],
                username=student2.name,
                firstname=student2.firstname,
                lastname=student2.lastname,
            ),
        ],
        teachers=[
            dict(
                user_id=teacher1.udm_properties["brokerID"],
                username=teacher1.name,
            )
        ],
    )
    assert response.json() == expected_group


@pytest.mark.asyncio
async def test_extract_members(monkeypatch):
    users = [
        teacher1,
        student1,
        student2,
    ]
    monkeypatch.setattr(dependencies, "get_single_user_by_name", AsyncMock(side_effect=users))
    classmembers = await dependencies.extract_members(
        SchoolClass(**class1), get_preconfigured_session_mock()
    )
    assert classmembers.dict() == dict(
        students=[
            dict(
                user_id=student1.udm_properties["brokerID"],
                username=student1.name,
                firstname=student1.firstname,
                lastname=student1.lastname,
            ),
            dict(
                user_id=student2.udm_properties["brokerID"],
                username=student2.name,
                firstname=student2.firstname,
                lastname=student2.lastname,
            ),
        ],
        teachers=[
            dict(
                user_id=teacher1.udm_properties["brokerID"],
                username=teacher1.name,
            )
        ],
    )


@pytest.mark.parametrize("user", [student1, teacher1])
def test_roles_in_school(user):
    roles = dependencies.roles_in_school(user, user.school)
    assert roles == set(user.roles)


@pytest.mark.asyncio
async def test_extract_members_filtered(monkeypatch, caplog):
    """Test if extract filters users without brokerID and logs a warning for them"""
    users = [student3, student4]
    monkeypatch.setattr(dependencies, "get_single_user_by_name", AsyncMock(side_effect=users))
    classmembers = await dependencies.extract_members(
        SchoolClass(**class2), get_preconfigured_session_mock()
    )
    assert classmembers.dict() == dict(
        students=[
            dict(
                user_id=student3.udm_properties["brokerID"],
                username=student3.name,
                firstname=student3.firstname,
                lastname=student3.lastname,
            ),
        ],
        teachers=[],
    )
    assert "udm_properties -> brokerID field required (type=value_error.missing)" in caplog.text


def test_user_groups_filtered(client, monkeypatch, caplog):
    class_members = [
        student3,
        student4,
    ]
    monkeypatch.setattr(
        dependencies,
        "get_user",
        AsyncMock(return_value=student3),
    )
    monkeypatch.setattr(
        dependencies,
        "get_single_user_by_name",
        AsyncMock(side_effect=class_members),
    )
    monkeypatch.setattr(
        dependencies,
        "schoolclass_getter",
        AsyncMock(return_value=class2),
    )
    response = client.get(f"/users/{student3.udm_properties['brokerID']}/groups")
    assert response.status_code == 200
    expected_groups = dict(
        groups=[
            {
                "group_id": f"{class2.school}-{class2.name}",
                "name": f"{class2.school}-{class2.name}",
                "student_count": 1,
            }
        ]
    )
    assert response.json() == expected_groups
    assert "udm_properties -> brokerID field required (type=value_error.missing)" in caplog.text


@pytest.mark.skip("Test does not work and need to be fixed when proper pseudonyms are implemented")
def test_group_users_filtered(client, monkeypatch, caplog):
    users = [student3, student4]
    monkeypatch.setattr(
        dependencies,
        "schoolclass_getter",
        AsyncMock(return_value=class2),
    )
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(side_effect=users))
    response = client.get(f"/groups/{class2.school}-{class2.name}/users")
    assert response.status_code == 200
    expected_group = dict(
        students=[
            dict(
                user_id=student3.udm_properties["brokerID"],
                username=student3.name,
                firstname=student3.firstname,
                lastname=student3.lastname,
            ),
        ],
        teachers=[],
    )
    assert response.json() == expected_group
    assert "udm_properties -> brokerID field required (type=value_error.missing)" in caplog.text
